@@ -1,6 +1,6 @@
 /* Code to manage a subscriber's MSC-A role */
 /*
- * (C) 2019 by sysmocom - s.m.f.c. GmbH <info@sysmocom.de>
+ * (C) 2019 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
  * All Rights Reserved
  *
  * SPDX-License-Identifier: AGPL-3.0+
@@ -36,6 +36,7 @@
 #include <osmocom/msc/signal.h>
 #include <osmocom/msc/vlr.h>
 #include <osmocom/msc/transaction.h>
+#include <osmocom/msc/transaction_cc.h>
 #include <osmocom/msc/ran_peer.h>
 #include <osmocom/msc/ran_msg_a.h>
 #include <osmocom/msc/ran_msg_iu.h>
@@ -46,6 +47,8 @@
 #include <osmocom/msc/call_leg.h>
 #include <osmocom/msc/rtp_stream.h>
 #include <osmocom/msc/msc_ho.h>
+#include <osmocom/msc/codec_mapping.h>
+#include <osmocom/msc/msc_vgcs.h>
 
 #define MSC_A_USE_WAIT_CLEAR_COMPLETE "wait-Clear-Complete"
 
@@ -486,6 +489,16 @@ static bool msc_a_fsm_has_active_transactions(struct osmo_fsm_inst *fi)
 			 __func__);
 		return true;
 	}
+	if (osmo_use_count_by(&msc_a->use_count, MSC_A_USE_CM_SERVICE_GCC)) {
+		LOG_MSC_A(msc_a, LOGL_DEBUG, "%s: still awaiting MO GCC request after a CM Service Request\n",
+			 __func__);
+		return true;
+	}
+	if (osmo_use_count_by(&msc_a->use_count, MSC_A_USE_CM_SERVICE_BCC)) {
+		LOG_MSC_A(msc_a, LOGL_DEBUG, "%s: still awaiting MO BCC request after a CM Service Request\n",
+			 __func__);
+		return true;
+	}
 	if (osmo_use_count_by(&msc_a->use_count, MSC_A_USE_CM_SERVICE_SMS)) {
 		LOG_MSC_A(msc_a, LOGL_DEBUG, "%s: still awaiting MO SMS after a CM Service Request\n",
 			 __func__);
@@ -562,6 +575,66 @@ static void msc_a_fsm_authenticated(struct osmo_fsm_inst *fi, uint32_t event, vo
 	}
 }
 
+static struct call_leg *msc_a_ensure_call_leg(struct msc_a *msc_a, struct gsm_trans *for_cc_trans)
+{
+	struct call_leg *cl = msc_a->cc.call_leg;
+	struct gsm_network *net = msc_a_net(msc_a);
+
+	/* Ensure that events about RTP endpoints coming from the msc_a->cc.call_leg know which gsm_trans to abort on
+	 * error */
+	if (!msc_a->cc.active_trans)
+		msc_a->cc.active_trans = for_cc_trans;
+	if (msc_a->cc.active_trans != for_cc_trans) {
+		LOG_TRANS(for_cc_trans, LOGL_ERROR,
+			  "Cannot create call leg, another trans is already active for this conn\n");
+		return NULL;
+	}
+
+	if (!cl) {
+		cl = msc_a->cc.call_leg = call_leg_alloc(msc_a->c.fi,
+							 MSC_EV_CALL_LEG_TERM,
+							 MSC_EV_CALL_LEG_RTP_LOCAL_ADDR_AVAILABLE,
+							 MSC_EV_CALL_LEG_RTP_COMPLETE);
+		OSMO_ASSERT(cl);
+
+		if (net->use_osmux != OSMUX_USAGE_OFF) {
+			struct msc_i *msc_i = msc_a_msc_i(msc_a);
+			if (msc_i->c.remote_to) {
+				/* TODO: investigate what to do in this case */
+				LOG_MSC_A(msc_a, LOGL_ERROR, "Osmux not yet supported for inter-MSC");
+			} else {
+				cl->ran_peer_supports_osmux = msc_i->ran_conn->ran_peer->remote_supports_osmux;
+			}
+		}
+
+	}
+	return cl;
+}
+
+int msc_a_ensure_cn_local_rtp(struct msc_a *msc_a, struct gsm_trans *cc_trans)
+{
+	struct call_leg *cl;
+	struct rtp_stream *rtp_to_ran;
+
+	cl = msc_a_ensure_call_leg(msc_a, cc_trans);
+	if (!cl)
+		return -EINVAL;
+	rtp_to_ran = cl->rtp[RTP_TO_RAN];
+
+	if (call_leg_local_ip(cl, RTP_TO_CN)) {
+		/* Already has an RTP address and port towards the CN, continue right away. */
+		return osmo_fsm_inst_dispatch(msc_a->c.fi, MSC_EV_CALL_LEG_RTP_LOCAL_ADDR_AVAILABLE, cl->rtp[RTP_TO_CN]);
+	}
+
+	/* No CN RTP address available yet, ask the MGW to create one.
+	 * Set a codec to be used: if Assignment on the RAN side is already done, take the same codec as the RTP_TO_RAN.
+	 * If no RAN side RTP is established, try to guess a preliminary codec from SDP -- before Assignment, picking a
+	 * codec from the SDP is more politeness/avoiding confusion than necessity. The actual codec to be used would be
+	 * determined later. If no codec could be determined, pass none for the time being. */
+	return call_leg_ensure_ci(cl, RTP_TO_CN, cc_trans->call_id, cc_trans,
+				  rtp_to_ran->codecs_known ? &rtp_to_ran->codecs : NULL, NULL);
+}
+
 /* The MGW has given us a local IP address for the RAN side. Ready to start the Assignment of a voice channel. */
 static void msc_a_call_leg_ran_local_addr_available(struct msc_a *msc_a)
 {
@@ -575,16 +648,51 @@ static void msc_a_call_leg_ran_local_addr_available(struct msc_a *msc_a)
 		return;
 	}
 
-	/* Once a CI is known, we could also CRCX the CN side of the MGW endpoint, but it makes sense to wait for the
-	 * codec to be determined by the Assignment Complete message, first. */
+	trans_cc_filter_run(cc_trans);
+	LOG_TRANS(cc_trans, LOGL_DEBUG, "Sending Assignment Command\n");
 
-	if (mncc_bearer_cap_to_channel_type(&channel_type, &cc_trans->bearer_cap)) {
-		LOG_MSC_A(msc_a, LOGL_ERROR, "Cannot compose Channel Type from bearer capabilities\n");
-		trans_free(cc_trans);
+	switch (cc_trans->bearer_cap.transfer) {
+	case GSM48_BCAP_ITCAP_SPEECH:
+		if (!cc_trans->cc.local.audio_codecs.count) {
+			LOG_TRANS(cc_trans, LOGL_ERROR, "Assignment not possible, no matching codec: %s\n",
+				  codec_filter_to_str(&cc_trans->cc.codecs, &cc_trans->cc.local, &cc_trans->cc.remote));
+			call_leg_release(msc_a->cc.call_leg);
+			return;
+		}
+
+		/* Compose 48.008 Channel Type from the current set of codecs
+		 * determined from both local and remote codec capabilities. */
+		if (sdp_audio_codecs_to_gsm0808_channel_type(&channel_type, &cc_trans->cc.local.audio_codecs)) {
+			LOG_MSC_A(msc_a, LOGL_ERROR, "Cannot compose Channel Type (Permitted Speech) from codecs: %s\n",
+				  codec_filter_to_str(&cc_trans->cc.codecs, &cc_trans->cc.local, &cc_trans->cc.remote));
+			trans_free(cc_trans);
+			return;
+		}
+		break;
+	case GSM48_BCAP_ITCAP_UNR_DIG_INF:
+		if (!cc_trans->cc.local.bearer_services.count) {
+			LOG_TRANS(cc_trans, LOGL_ERROR, "Assignment not possible, no matching bearer service: %s\n",
+				  csd_filter_to_str(&cc_trans->cc.csd, &cc_trans->cc.local, &cc_trans->cc.remote));
+			call_leg_release(msc_a->cc.call_leg);
+			return;
+		}
+
+		/* Compose 48.008 Channel Type from the current set of bearer
+		 * services determined from local and remote capabilities. */
+		if (csd_bs_list_to_gsm0808_channel_type(&channel_type, &cc_trans->cc.local.bearer_services)) {
+			LOG_MSC_A(msc_a, LOGL_ERROR, "Cannot compose channel type from: %s\n",
+				  csd_filter_to_str(&cc_trans->cc.csd, &cc_trans->cc.local, &cc_trans->cc.remote));
+			return;
+		}
+		break;
+	default:
+		LOG_TRANS(cc_trans, LOGL_ERROR, "Assignment not possible for information transfer capability %d\n",
+			  cc_trans->bearer_cap.transfer);
+		call_leg_release(msc_a->cc.call_leg);
 		return;
 	}
 
-	/* The RAN side RTP address is known, so the voice Assignment can commence. */
+	/* The RAN side RTP address is known, so the voice/CSD Assignment can commence. */
 	msg = (struct ran_msg){
 		.msg_type = RAN_MSG_ASSIGNMENT_COMMAND,
 		.assignment_command = {
@@ -593,7 +701,7 @@ static void msc_a_call_leg_ran_local_addr_available(struct msc_a *msc_a)
 			.osmux_present = msc_a->cc.call_leg->rtp[RTP_TO_RAN]->use_osmux,
 			.osmux_cid = msc_a->cc.call_leg->rtp[RTP_TO_RAN]->local_osmux_cid,
 			.call_id_present = true,
-			.call_id = cc_trans->callref,
+			.call_id = cc_trans->call_id,
 			.lcls = cc_trans->cc.lcls,
 		},
 	};
@@ -604,15 +712,6 @@ static void msc_a_call_leg_ran_local_addr_available(struct msc_a *msc_a)
 	}
 }
 
-static void msc_a_call_leg_cn_local_addr_available(struct msc_a *msc_a, struct gsm_trans *cc_trans)
-{
-	if (gsm48_tch_rtp_create(cc_trans)) {
-		LOG_MSC_A(msc_a, LOGL_ERROR, "Cannot inform MNCC of RTP address\n");
-		trans_free(cc_trans);
-		return;
-	}
-}
-
 static struct gsm_trans *find_waiting_call(struct msc_a *msc_a)
 {
 	struct gsm_trans *trans;
@@ -706,7 +805,7 @@ static void msc_a_fsm_communicating(struct osmo_fsm_inst *fi, uint32_t event, vo
 			msc_a_call_leg_ran_local_addr_available(msc_a);
 			return;
 		case RTP_TO_CN:
-			msc_a_call_leg_cn_local_addr_available(msc_a, rtps->for_trans);
+			cc_on_cn_local_rtp_port_known(rtps->for_trans);
 			return;
 		default:
 			LOG_MSC_A(msc_a, LOGL_ERROR, "Invalid data for %s\n", osmo_fsm_event_name(fi->fsm, event));
@@ -785,6 +884,8 @@ static void msc_a_fsm_releasing_onenter(struct osmo_fsm_inst *fi, uint32_t prev_
 		MSC_A_USE_CM_SERVICE_CC,
 		MSC_A_USE_CM_SERVICE_SMS,
 		MSC_A_USE_CM_SERVICE_SS,
+		MSC_A_USE_CM_SERVICE_GCC,
+		MSC_A_USE_CM_SERVICE_BCC,
 		MSC_A_USE_PAGING_RESPONSE,
 	};
 
@@ -1285,6 +1386,10 @@ int msc_a_up_l3(struct msc_a *msc_a, struct msgb *msg)
 #endif
 
 	switch (pdisc) {
+	case GSM48_PDISC_GROUP_CC:
+	case GSM48_PDISC_BCAST_CC:
+		rc = gsm44068_rcv_bcc_gcc(msc_a, NULL, msg);
+		break;
 	case GSM48_PDISC_CC:
 		rc = gsm0408_rcv_cc(msc_a, msg);
 		break;
@@ -1321,10 +1426,22 @@ int msc_a_up_l3(struct msc_a *msc_a, struct msgb *msg)
 
 static void msc_a_up_call_assignment_complete(struct msc_a *msc_a, const struct ran_msg *ac)
 {
-	struct gsm_trans *cc_trans = msc_a->cc.active_trans;
+	struct gsm_trans *cc_trans = msc_a->cc.active_trans, *gcc_trans;
 	struct rtp_stream *rtps_to_ran = msc_a->cc.call_leg ? msc_a->cc.call_leg->rtp[RTP_TO_RAN] : NULL;
-	const enum mgcp_codecs *codec_if_known = ac->assignment_complete.codec_present ?
-							&ac->assignment_complete.codec : NULL;
+	const struct gsm0808_speech_codec *codec_if_known = ac->assignment_complete.codec_present ?
+							    &ac->assignment_complete.codec : NULL;
+
+	/* For a voice group call, handling is performed by VGCS FSM */
+	gcc_trans = trans_find_by_type(msc_a, TRANS_GCC);
+	if (gcc_trans) {
+		vgcs_vbs_caller_assign_cpl(gcc_trans);
+		return;
+	}
+	gcc_trans = trans_find_by_type(msc_a, TRANS_BCC);
+	if (gcc_trans) {
+		vgcs_vbs_caller_assign_cpl(gcc_trans);
+		return;
+	}
 
 	if (!rtps_to_ran) {
 		LOG_MSC_A(msc_a, LOGL_ERROR, "Rx Assignment Complete, but no RTP stream is set up\n");
@@ -1342,25 +1459,69 @@ static void msc_a_up_call_assignment_complete(struct msc_a *msc_a, const struct
 		return;
 	}
 
-	/* Update RAN-side endpoint CI: */
-	if (codec_if_known)
-		rtp_stream_set_codec(rtps_to_ran, *codec_if_known);
+	if (codec_if_known) {
+		const struct codec_mapping *codec_cn;
+		/* For 2G:
+		 * - The Assignment Complete has returned a specific codec (e.g. FR3 for AMR FR).
+		 * - Set this codec at the MGW endpoint facing the RAN.
+		 * - Also set this codec at the MGW endpoint facing the CN -- we require an exact match on both call
+		 *   legs.
+		 *   - TODO: be aware of transcoding that the MGW is capable of, e.g. AMR octet-aligned to AMR
+		 *     bandwidth-efficient...
+		 *
+		 * For 3G:
+		 * - ran_infra->force_mgw_codecs_to_ran sets VND.3GPP.IUFP as single codec at the MGW towards RAN.
+		 * - ran_msg_iu.c always returns FR3 (AMR FR) for the assigned codec. Set that at the MGW towards CN.
+		 * - So the MGW decapsulates IuUP <-> AMR
+		 */
+		codec_cn = codec_mapping_by_gsm0808_speech_codec_type(codec_if_known->type);
+		/* TODO: use codec_mapping_by_gsm0808_speech_codec() to also match on codec_if_known->cfg */
+		if (!codec_cn) {
+			LOG_TRANS(cc_trans, LOGL_ERROR, "Unknown codec in Assignment Complete: %s\n",
+				  gsm0808_speech_codec_type_name(codec_if_known->type));
+			call_leg_release(msc_a->cc.call_leg);
+			return;
+		}
+
+		/* Check for unexpected codec with CSD */
+		if (cc_trans->bearer_cap.transfer == GSM48_BCAP_ITCAP_UNR_DIG_INF &&
+		    codec_if_known->type != GSM0808_SCT_CSD) {
+			LOG_TRANS(cc_trans, LOGL_ERROR, "Unexpected codec in Assignment Complete for CSD: %s\n",
+				  gsm0808_speech_codec_type_name(codec_if_known->type));
+			call_leg_release(msc_a->cc.call_leg);
+			return;
+		}
+
+		/* Update RAN-side endpoint CI from Assignment result -- unless it is forced by the ran_infra, in which
+		 * case it remains unchanged as passed to the earlier call of call_leg_ensure_ci(). */
+		if (msc_a->c.ran->force_mgw_codecs_to_ran.count == 0)
+			rtp_stream_set_one_codec(rtps_to_ran, &codec_cn->sdp);
+
+		/* Update codec filter with Assignment result, for the CN side */
+		cc_trans->cc.codecs.assignment = codec_cn->sdp;
+	} else {
+		/* No codec passed in Assignment Complete, set 'codecs.assignment' to none. */
+		cc_trans->cc.codecs.assignment = (struct sdp_audio_codec){};
+		LOG_TRANS(cc_trans, LOGL_INFO, "Assignment Complete without voice codec\n");
+	}
+
 	rtp_stream_set_remote_addr(rtps_to_ran, &ac->assignment_complete.remote_rtp);
 	if (rtps_to_ran->use_osmux)
 		rtp_stream_set_remote_osmux_cid(rtps_to_ran,
 						ac->assignment_complete.osmux_cid);
-
 	rtp_stream_commit(rtps_to_ran);
 
-	/* Setup CN side endpoint CI:
-	 * Now that
-	 * - the first CI has been created and a definitive endpoint name is assigned to the call_leg's MGW
-	 *   endpoint,
-	 * - the Assignment has chosen a speech codec
-	 * go on to create the CN side RTP stream's CI. */
-	if (call_leg_ensure_ci(msc_a->cc.call_leg, RTP_TO_CN, cc_trans->callref, cc_trans,
-			       codec_if_known, NULL)) {
-		LOG_MSC_A_CAT(msc_a, DCC, LOGL_ERROR, "Error creating MGW CI towards CN\n");
+	/* Remember the Codec List (BSS Supported) */
+	if (ac->assignment_complete.codec_list_bss_supported)
+		codec_filter_set_bss(&cc_trans->cc.codecs, ac->assignment_complete.codec_list_bss_supported);
+
+	trans_cc_filter_run(cc_trans);
+	LOG_TRANS(cc_trans, LOGL_INFO, "Assignment Complete: RAN: %s, CN: %s\n",
+		  sdp_audio_codecs_to_str(&rtps_to_ran->codecs),
+		  sdp_audio_codecs_to_str(&cc_trans->cc.local.audio_codecs));
+
+	if (cc_on_assignment_done(cc_trans)) {
+		/* If an error occurred, it was logged in cc_assignment_done() */
 		call_leg_release(msc_a->cc.call_leg);
 		return;
 	}
@@ -1377,6 +1538,18 @@ static void msc_a_up_call_assignment_failure(struct msc_a *msc_a, const struct r
 		return;
 	}
 
+	/* For a voice group call, release is performed by VGCS FSM */
+	trans = trans_find_by_type(msc_a, TRANS_GCC);
+	if (trans) {
+		vgcs_vbs_caller_assign_fail(trans);
+		return;
+	}
+	trans = trans_find_by_type(msc_a, TRANS_BCC);
+	if (trans) {
+		vgcs_vbs_caller_assign_fail(trans);
+		return;
+	}
+
 	/* Otherwise, a silent call might be active */
 	trans = trans_find_by_type(msc_a, TRANS_SILENT_CALL);
 	if (trans) {
@@ -1442,6 +1615,20 @@ int msc_a_ran_dec_from_msc_i(struct msc_a *msc_a, struct msc_a_ran_dec_data *d)
 			.lai.plmn = msc_a_net(msc_a)->plmn,
 		};
 		gsm0808_cell_id_to_cgi(&msc_a->via_cell, msg->compl_l3.cell_id);
+
+		/* If a codec list was sent along in the RAN_MSG_COMPL_L3, remember it for any upcoming codec
+		 * resolution. */
+		if (msg->compl_l3.codec_list_bss_supported) {
+			msc_a->cc.compl_l3_codec_list_bss_supported = *msg->compl_l3.codec_list_bss_supported;
+			if (log_check_level(msc_a->c.ran->log_subsys, LOGL_DEBUG)) {
+				struct sdp_audio_codecs ac = {};
+				sdp_audio_codecs_from_speech_codec_list(&ac, &msc_a->cc.compl_l3_codec_list_bss_supported);
+				LOG_MSC_A(msc_a, LOGL_DEBUG, "Complete Layer 3: Codec List (BSS Supported): %s\n",
+					  sdp_audio_codecs_to_str(&ac));
+			}
+		}
+
+		/* Submit the Complete Layer 3 Information DTAP */
 		rc = msc_a_up_l3(msc_a, msg->compl_l3.msg);
 		if (!rc) {
 			struct ran_conn *conn = msub_ran_conn(msc_a->c.msub);
@@ -1569,6 +1756,136 @@ int msc_a_ran_dec_from_msc_i(struct msc_a *msc_a, struct msc_a_ran_dec_data *d)
 	return rc;
 }
 
+static int msc_a_rx_vgcs_bss_decoded(struct osmo_fsm_inst *caller_fi, void *caller_data, const struct ran_msg *msg)
+{
+	struct vgcs_bss *bss = caller_data;
+	struct msc_a *msc_a = (bss->trans) ? bss->trans->msc_a : NULL;
+	int rc = 0;
+
+	switch (msg->msg_type) {
+	case RAN_MSG_VGCS_VBS_SETUP_ACK:
+		/* The BSS accepts VGCS/VBS and sends us supported features. */
+		vgcs_vbs_setup_ack(bss, msg);
+		break;
+	case RAN_MSG_VGCS_VBS_SETUP_REFUSE:
+		/* The BSS refuses VGCS/VBS. */
+		vgcs_vbs_setup_refuse(bss, msg);
+		break;
+	case RAN_MSG_UPLINK_REQUEST:
+		/* A mobile station requests the uplink on a VGCS channel. */
+		vgcs_uplink_request(bss, msg);
+		break;
+	case RAN_MSG_UPLINK_REQUEST_CNF:
+		/* The uplink on a VGCS channel has been established. */
+		vgcs_uplink_request_cnf(bss, msg);
+		break;
+	case RAN_MSG_UPLINK_APPLICATION_DATA:
+		/* Application data received on the uplink of a VGCS channel. */
+		vgcs_app_data(bss, msg);
+		break;
+	case RAN_MSG_DTAP:
+		/* BSS confirms the release of the channel. */
+		vgcs_bss_dtap(bss, msg);
+		break;
+	case RAN_MSG_UPLINK_RELEASE_IND:
+		/* A mobile station releases the uplink on a VGCS channel. */
+		vgcs_uplink_release_ind(bss, msg);
+		break;
+	case RAN_MSG_CLEAR_REQUEST:
+		/* BSS indicated that the channel has been released. */
+		vgcs_vbs_clear_req(bss, msg);
+		break;
+	case RAN_MSG_CLEAR_COMPLETE:
+		/* BSS confirms the release of the channel. */
+		vgcs_vbs_clear_cpl(bss, msg);
+		break;
+	default:
+		LOG_MSC_A(msc_a, LOGL_ERROR, "VGCS message from BSS not implemented: %s\n",
+			  ran_msg_type_name(msg->msg_type));
+		rc = -ENOTSUP;
+		break;
+	}
+	return rc;
+}
+
+int msc_a_rx_vgcs_bss(struct vgcs_bss *bss, struct ran_conn *from_conn, struct msgb *msg)
+{
+	struct ran_dec ran_dec;
+
+	/* Feed through the decoding mechanism ran_msg. The decoded message arrives in msc_a_rx_vgcs_decoded() */
+	ran_dec = (struct ran_dec) {
+		.caller_data = bss,
+		.decode_cb = msc_a_rx_vgcs_bss_decoded,
+	};
+	struct ran_peer *ran_peer = from_conn->ran_peer;
+	struct ran_infra *ran = ran_peer->sri->ran;
+	if (!ran->ran_dec_l2) {
+		LOGP(DMSC, LOGL_ERROR, "No ran_dec_l2() defined for RAN type %s\n",
+		     osmo_rat_type_name(ran->type));
+		return -ENOTSUP;
+	}
+	return ran->ran_dec_l2(&ran_dec, msg);
+}
+
+static int msc_a_rx_vgcs_cell_decoded(struct osmo_fsm_inst *caller_fi, void *caller_data, const struct ran_msg *msg)
+{
+	struct vgcs_bss_cell *cell = caller_data;
+	struct msc_a *msc_a = (cell->bss && cell->bss->trans) ? cell->bss->trans->msc_a : NULL;
+	int rc = 0;
+
+	switch (msg->msg_type) {
+	case RAN_MSG_VGCS_VBS_ASSIGN_RES:
+		/* The BSS accepts VGCS/VBS channel assignment. */
+		vgcs_vbs_assign_result(cell, msg);
+		break;
+	case RAN_MSG_VGCS_VBS_ASSIGN_FAIL:
+		/* The BSS refuses VGCS/VBS channel assignment. */
+		vgcs_vbs_assign_fail(cell, msg);
+		break;
+	case RAN_MSG_VGCS_VBS_QUEUING_IND:
+		/* The BSS needs more time for VGCS/VBS channel assignment. */
+		vgcs_vbs_queuing_ind(cell);
+		break;
+	case RAN_MSG_VGCS_VBS_ASSIGN_STATUS:
+		/* The BSS gives cell status about VGCS/VBS channel. */
+		vgcs_vbs_assign_status(cell, msg);
+		break;
+	case RAN_MSG_CLEAR_REQUEST:
+		/* BSS indicated that the channel has been released. */
+		vgcs_vbs_clear_req_channel(cell, msg);
+		break;
+	case RAN_MSG_CLEAR_COMPLETE:
+		/* BSS confirms the release of the channel. */
+		vgcs_vbs_clear_cpl_channel(cell, msg);
+		break;
+	default:
+		LOG_MSC_A(msc_a, LOGL_ERROR, "Message from BSS leg not implemented: %s\n",
+			  ran_msg_type_name(msg->msg_type));
+		rc = -ENOTSUP;
+		break;
+	}
+	return rc;
+}
+
+int msc_a_rx_vgcs_cell(struct vgcs_bss_cell *cell, struct ran_conn *from_conn, struct msgb *msg)
+{
+	struct ran_dec ran_dec;
+
+	/* Feed through the decoding mechanism ran_msg. The decoded message arrives in msc_a_rx_vgcs_decoded() */
+	ran_dec = (struct ran_dec) {
+		.caller_data = cell,
+		.decode_cb = msc_a_rx_vgcs_cell_decoded,
+	};
+	struct ran_peer *ran_peer = from_conn->ran_peer;
+	struct ran_infra *ran = ran_peer->sri->ran;
+	if (!ran->ran_dec_l2) {
+		LOGP(DMSC, LOGL_ERROR, "No ran_dec_l2() defined for RAN type %s\n",
+		     osmo_rat_type_name(ran->type));
+		return -ENOTSUP;
+	}
+	return ran->ran_dec_l2(&ran_dec, msg);
+}
+
 static int msc_a_ran_dec_from_msc_t(struct msc_a *msc_a, struct msc_a_ran_dec_data *d)
 {
 	struct msc_t *msc_t = msc_a_msc_t(msc_a);
@@ -1737,52 +2054,49 @@ int msc_tx_common_id(struct msc_a *msc_a, enum msc_role to_role)
 
 static int msc_a_start_assignment(struct msc_a *msc_a, struct gsm_trans *cc_trans)
 {
-	struct call_leg *cl = msc_a->cc.call_leg;
-	struct msc_i *msc_i = msc_a_msc_i(msc_a);
-	struct gsm_network *net = msc_a_net(msc_a);
-	enum mgcp_codecs codec, *codec_ptr;
+	struct call_leg *cl;
+	bool cn_rtp_available;
+	bool ran_rtp_available;
 
 	OSMO_ASSERT(!msc_a->cc.active_trans);
 	msc_a->cc.active_trans = cc_trans;
 
+	cc_trans->cc.codecs.assignment = (struct sdp_audio_codec){};
+
 	OSMO_ASSERT(cc_trans && cc_trans->type == TRANS_CC);
+	cl = msc_a_ensure_call_leg(msc_a, cc_trans);
+	if (!cl)
+		return -EINVAL;
 
-	if (!cl) {
-		cl = msc_a->cc.call_leg = call_leg_alloc(msc_a->c.fi,
-							 MSC_EV_CALL_LEG_TERM,
-							 MSC_EV_CALL_LEG_RTP_LOCAL_ADDR_AVAILABLE,
-							 MSC_EV_CALL_LEG_RTP_COMPLETE);
-		OSMO_ASSERT(cl);
+	/* See if we can set a preliminary codec. If not, pass none for the time being. */
+	trans_cc_filter_run(cc_trans);
 
-		/* HACK: We put the connection in loopback mode at the beginning to
-		 * trick the hNodeB into doing the IuUP negotiation with itself.
-		 * This is a hack we need because osmo-mgw does not support IuUP yet, see OS#2459. */
-		if (msc_a->c.ran->type == OSMO_RAT_UTRAN_IU)
-			cl->crcx_conn_mode[RTP_TO_RAN] = MGCP_CONN_LOOPBACK;
+	cn_rtp_available = call_leg_local_ip(cl, RTP_TO_CN);
+	ran_rtp_available = call_leg_local_ip(cl, RTP_TO_RAN);
+
+	/* Set up RTP ports for both RAN and CN side. Even though we ask for both at the same time, the
+	 * osmo_mgcpc_ep_fsm automagically waits for the first CRCX to complete before firing the second CRCX. The one
+	 * issued first here will also be the first CRCX sent to the MGW. Usually both still need to be set up. */
+	if (!cn_rtp_available)
+		call_leg_ensure_ci(cl, RTP_TO_CN, cc_trans->call_id, cc_trans,
+				   &cc_trans->cc.local.audio_codecs, NULL);
+	if (!ran_rtp_available) {
+		struct sdp_audio_codecs *codecs;
+		if (msc_a->c.ran->force_mgw_codecs_to_ran.count)
+			codecs = &msc_a->c.ran->force_mgw_codecs_to_ran;
+		else
+			codecs = &cc_trans->cc.local.audio_codecs;
+		return call_leg_ensure_ci(cl, RTP_TO_RAN, cc_trans->call_id, cc_trans, codecs, NULL);
 	}
 
-	if (net->use_osmux != OSMUX_USAGE_OFF) {
-		msc_i = msc_a_msc_i(msc_a);
-		if (msc_i->c.remote_to) {
-			/* TODO: investigate what to do in this case */
-			LOG_MSC_A(msc_a, LOGL_ERROR, "Osmux not yet supported for inter-MSC");
-		} else {
-			cl->ran_peer_supports_osmux = msc_i->ran_conn->ran_peer->remote_supports_osmux;
-		}
-	}
-
-	/* This will lead to either MSC_EV_CALL_LEG_LOCAL_ADDR_AVAILABLE or MSC_EV_CALL_LEG_TERM.
-	 * If the local address is already known, then immediately trigger. */
-	if (call_leg_local_ip(cl, RTP_TO_RAN))
+	/* Should these already be set up, immediately continue by retriggering the events signalling that the RTP
+	 * ports are available. The ordering is: first CN, then RAN. */
+	if (cn_rtp_available && ran_rtp_available)
 		return osmo_fsm_inst_dispatch(msc_a->c.fi, MSC_EV_CALL_LEG_RTP_LOCAL_ADDR_AVAILABLE, cl->rtp[RTP_TO_RAN]);
-
-	if (msc_a->c.ran->type == OSMO_RAT_UTRAN_IU) {
-		codec = CODEC_IUFP;
-		codec_ptr = &codec;
-	} else {
-		codec_ptr = NULL;
-	}
-	return call_leg_ensure_ci(msc_a->cc.call_leg, RTP_TO_RAN, cc_trans->callref, cc_trans, codec_ptr, NULL);
+	else if (cn_rtp_available)
+		return osmo_fsm_inst_dispatch(msc_a->c.fi, MSC_EV_CALL_LEG_RTP_LOCAL_ADDR_AVAILABLE, cl->rtp[RTP_TO_CN]);
+	/* Otherwise wait for MGCP response and continue from there.  */
+	return 0;
 }
 
 int msc_a_try_call_assignment(struct gsm_trans *cc_trans)
@@ -1808,8 +2122,10 @@ int msc_a_try_call_assignment(struct gsm_trans *cc_trans)
  * Given a CM Service type, return a matching token intended for osmo_use_count.
  * For unknown service type, return NULL.
  */
-const char *msc_a_cm_service_type_to_use(enum osmo_cm_service_type cm_service_type)
+const char *msc_a_cm_service_type_to_use(struct msc_a *msc_a, enum osmo_cm_service_type cm_service_type)
 {
+	struct gsm_network *net = msc_a_net(msc_a);
+
 	switch (cm_service_type) {
 	case GSM48_CMSERV_MO_CALL_PACKET:
 	case GSM48_CMSERV_EMERGENCY:
@@ -1821,6 +2137,18 @@ const char *msc_a_cm_service_type_to_use(enum osmo_cm_service_type cm_service_ty
 	case GSM48_CMSERV_SUP_SERV:
 		return MSC_A_USE_CM_SERVICE_SS;
 
+	case GSM48_CMSERV_VGCS:
+		if (net->asci.enable)
+			return MSC_A_USE_CM_SERVICE_GCC;
+		else
+			return NULL;
+
+	case GSM48_CMSERV_VBS:
+		if (net->asci.enable)
+			return MSC_A_USE_CM_SERVICE_BCC;
+		else
+			return NULL;
+
 	default:
 		return NULL;
 	}
