@@ -1,6 +1,6 @@
 /* MSC Handover implementation */
 /*
- * (C) 2019 by sysmocom - s.m.f.c. GmbH <info@sysmocom.de>
+ * (C) 2019 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
  * All Rights Reserved
  *
  * Author: Neels Hofmeyr
@@ -16,10 +16,6 @@
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 #include <osmocom/core/fsm.h>
@@ -43,6 +39,7 @@
 #include <osmocom/msc/call_leg.h>
 #include <osmocom/msc/rtp_stream.h>
 #include <osmocom/msc/mncc_call.h>
+#include <osmocom/msc/codec_mapping.h>
 
 struct osmo_fsm msc_ho_fsm;
 
@@ -383,6 +380,7 @@ static void msc_ho_send_handover_request(struct msc_a *msc_a)
 	struct vlr_subscr *vsub = msc_a_vsub(msc_a);
 	struct gsm_network *net = msc_a_net(msc_a);
 	struct gsm0808_channel_type channel_type;
+	struct gsm0808_speech_codec_list scl;
 	struct gsm_trans *cc_trans = msc_a->cc.active_trans;
 	struct ran_msg ran_enc_msg = {
 		.msg_type = RAN_MSG_HANDOVER_REQUEST,
@@ -403,8 +401,6 @@ static void msc_ho_send_handover_request(struct msc_a *msc_a)
 			/* Don't send AoIP Transport Layer Address for inter-MSC Handover */
 			.rtp_ran_local = (msc_a->ho.new_cell.type == MSC_NEIGHBOR_TYPE_LOCAL_RAN_PEER)
 				? call_leg_local_ip(msc_a->cc.call_leg, RTP_TO_RAN) : NULL,
-			.call_id_present = true,
-			.call_id = cc_trans->callref,
 		},
 	};
 
@@ -416,13 +412,41 @@ static void msc_ho_send_handover_request(struct msc_a *msc_a)
 			    osmo_hexdump_nospc_c(OTC_SELECT, msc_a->geran_encr.kc128, sizeof(msc_a->geran_encr.kc128))
 			    : "-");
 
-	if (msc_a->cc.active_trans) {
-		if (mncc_bearer_cap_to_channel_type(&channel_type, &msc_a->cc.active_trans->bearer_cap)) {
-			msc_ho_failed(msc_a, GSM0808_CAUSE_EQUIPMENT_FAILURE,
-				      "Failed to encode Bearer Cap to Channel Type\n");
+	if (cc_trans) {
+		switch (cc_trans->bearer_cap.transfer) {
+		case GSM48_BCAP_ITCAP_SPEECH:
+			if (sdp_audio_codecs_to_gsm0808_channel_type(&channel_type,
+								     &cc_trans->cc.local.audio_codecs)) {
+				msc_ho_failed(msc_a, GSM0808_CAUSE_EQUIPMENT_FAILURE,
+					      "Failed to determine Channel Type for Handover Request message (speech)\n");
+				return;
+			}
+			break;
+		case GSM48_BCAP_ITCAP_UNR_DIG_INF:
+			if (csd_bs_list_to_gsm0808_channel_type(&channel_type, &cc_trans->cc.local.bearer_services)) {
+				msc_ho_failed(msc_a, GSM0808_CAUSE_EQUIPMENT_FAILURE,
+					      "Failed to determine Channel Type for Handover Request message (CSD)\n");
+				return;
+			}
+			break;
+		default:
+			msc_ho_failed(msc_a, GSM0808_CAUSE_EQUIPMENT_FAILURE, "Failed to create"
+				      " Handover Request message for information transfer capability %d\n",
+				      cc_trans->bearer_cap.transfer);
 			return;
 		}
+
 		ran_enc_msg.handover_request.geran.channel_type = &channel_type;
+		ran_enc_msg.handover_request.call_id_present = true;
+		ran_enc_msg.handover_request.call_id = cc_trans->call_id;
+
+		sdp_audio_codecs_to_speech_codec_list(&scl, &cc_trans->cc.local.audio_codecs);
+		if (!scl.len) {
+			msc_ho_failed(msc_a, GSM0808_CAUSE_EQUIPMENT_FAILURE, "Failed to compose"
+				      " Codec List (MSC Preferred) for Handover Request message\n");
+			return;
+		}
+		ran_enc_msg.handover_request.codec_list_msc_preferred = &scl;
 	}
 
 	gsm0808_cell_id_from_cgi(&ran_enc_msg.handover_request.cell_id_serving, CELL_IDENT_WHOLE_GLOBAL, &vsub->cgi);
@@ -574,7 +598,7 @@ static int msc_ho_start_inter_msc_call_forwarding(struct msc_a *msc_a, struct ms
 
 	/* Backup old cell's RTP IP:port and codec data */
 	msc_a->ho.old_cell.ran_remote_rtp = rtp_to_ran->remote;
-	msc_a->ho.old_cell.codec = rtp_to_ran->codec;
+	msc_a->ho.old_cell.codecs = rtp_to_ran->codecs;
 
 	/* Blindly taken over from an MNCC trace of existing code: send an all-zero CCCAP: */
 	outgoing_call_req.fields |= MNCC_F_CCCAP;
@@ -684,7 +708,7 @@ static void msc_ho_rx_request_ack(struct msc_a *msc_a, struct msc_a_ran_dec_data
 	msc_a->ho.new_cell.codec = hra->ran_dec->handover_request_ack.codec;
 	if (hra->ran_dec->handover_request_ack.codec_present) {
 		LOG_HO(msc_a, LOGL_DEBUG, "Request Ack contains codec %s\n",
-		       osmo_mgcpc_codec_name(msc_a->ho.new_cell.codec));
+		       gsm0808_speech_codec_type_name(msc_a->ho.new_cell.codec.type));
 	}
 }
 
@@ -711,7 +735,7 @@ static void msc_ho_rtp_switch_to_new_cell(struct msc_a *msc_a)
 
 	/* Backup old cell's RTP IP:port and codec data */
 	msc_a->ho.old_cell.ran_remote_rtp = rtp_to_ran->remote;
-	msc_a->ho.old_cell.codec = rtp_to_ran->codec;
+	msc_a->ho.old_cell.codecs = rtp_to_ran->codecs;
 
 	LOG_HO(msc_a, LOGL_DEBUG, "Switching RTP stream to new cell: from " OSMO_SOCKADDR_STR_FMT " to " OSMO_SOCKADDR_STR_FMT "\n",
 	       OSMO_SOCKADDR_STR_FMT_ARGS(&msc_a->ho.old_cell.ran_remote_rtp),
@@ -730,10 +754,18 @@ static void msc_ho_rtp_switch_to_new_cell(struct msc_a *msc_a)
 
 	/* Switch over to the new peer */
 	rtp_stream_set_remote_addr(rtp_to_ran, &msc_a->ho.new_cell.ran_remote_rtp);
-	if (msc_a->ho.new_cell.codec_present)
-		rtp_stream_set_codec(rtp_to_ran, msc_a->ho.new_cell.codec);
-	else
+	if (msc_a->ho.new_cell.codec_present) {
+		const struct codec_mapping *m;
+		m = codec_mapping_by_gsm0808_speech_codec_type(msc_a->ho.new_cell.codec.type);
+		/* TODO: use codec_mapping_by_gsm0808_speech_codec() to also match on codec.cfg */
+		if (!m)
+			LOG_HO(msc_a, LOGL_ERROR, "Cannot resolve codec: %s\n",
+			       gsm0808_speech_codec_type_name(msc_a->ho.new_cell.codec.type));
+		else
+			rtp_stream_set_one_codec(rtp_to_ran, &m->sdp);
+	} else {
 		LOG_HO(msc_a, LOGL_ERROR, "No codec is set\n");
+	}
 	rtp_stream_commit(rtp_to_ran);
 }
 
@@ -772,7 +804,7 @@ static void msc_ho_rtp_rollback_to_old_cell(struct msc_a *msc_a)
 
 	/* Switch back to the old cell */
 	rtp_stream_set_remote_addr(rtp_to_ran, &msc_a->ho.old_cell.ran_remote_rtp);
-	rtp_stream_set_codec(rtp_to_ran, msc_a->ho.old_cell.codec);
+	rtp_stream_set_codecs(rtp_to_ran, &msc_a->ho.old_cell.codecs);
 	rtp_stream_commit(rtp_to_ran);
 }
 
