@@ -24,6 +24,7 @@
 #include "config.h"
 
 #include <osmocom/core/tdef.h>
+#include <osmocom/crypt/utran_cipher.h>
 
 #include <osmocom/msc/gsm_data.h>
 #include <osmocom/msc/vlr.h>
@@ -61,14 +62,12 @@ struct gsm_network *gsm_network_init(void *ctx, mncc_recv_cb_t mncc_recv)
 	if (!net)
 		return NULL;
 
-	net->sms_db_file_path = talloc_strdup(net, SMS_DEFAULT_DB_FILE_PATH);
-
 	net->plmn = (struct osmo_plmn_id){ .mcc=1, .mnc=1 };
 
 	/* Permit a compile-time default of A5/3 and A5/1 */
 	net->a5_encryption_mask = (1 << 3) | (1 << 1);
 	/* Permit a compile-time default of UEA2 and UEA1 */
-	net->uea_encryption_mask = (1 << 2) | (1 << 1);
+	net->uea_encryption_mask = (1 << OSMO_UTRAN_UEA2) | (1 << OSMO_UTRAN_UEA1);
 
 	net->mncc_guard_timeout = 180;
 	net->ncss_guard_timeout = 30;
@@ -76,6 +75,7 @@ struct gsm_network *gsm_network_init(void *ctx, mncc_recv_cb_t mncc_recv)
 	INIT_LLIST_HEAD(&net->trans_list);
 	INIT_LLIST_HEAD(&net->upqueue);
 	INIT_LLIST_HEAD(&net->neighbor_ident_list);
+	INIT_LLIST_HEAD(&net->asci.gcr_lists);
 
 	/* init statistics */
 	net->msc_ctrs = rate_ctr_group_alloc(net, &msc_ctrg_desc, 0);
