@@ -257,7 +257,7 @@ struct gsm_network *msc_network_alloc(void *ctx,
 	net->gsup_server_port = MSC_HLR_REMOTE_PORT_DEFAULT;
 
 	net->mgw.mgw_pool = mgcp_client_pool_alloc(net);
-	mgcp_client_conf_init(&net->mgw.conf);
+	net->mgw.conf = mgcp_client_conf_alloc(net);
 	net->call_waiting = true;
 	net->lcls_permitted = false;
 
@@ -330,9 +330,19 @@ static int msc_vty_go_parent(struct vty *vty)
 	case SMPP_NODE:
 	case MSC_NODE:
 	case MNCC_INT_NODE:
+	case ASCI_NODE:
 		vty->node = CONFIG_NODE;
 		vty->index = NULL;
 		break;
+	case GCR_NODE:
+		vty->node = ASCI_NODE;
+		vty->index = NULL;
+		break;
+	case VGC_NODE:
+	case VBC_NODE:
+		vty->node = GCR_NODE;
+		vty->index = NULL;
+		break;
 	case SUBSCR_NODE:
 		vty->node = ENABLE_NODE;
 		vty->index = NULL;
@@ -439,6 +449,18 @@ static const struct log_info_cat msc_default_categories[] = {
 		.color = "\033[1;32m",
 		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
+	[DBCC] = {
+		.name = "DBCC",
+		.description = "Layer3 Broadcast Call Control (BCC)",
+		.color = "\033[1;32m",
+		.enabled = 1, .loglevel = LOGL_NOTICE,
+	},
+	[DGCC] = {
+		.name = "DGCC",
+		.description = "Layer3 Group Call Control (GCC)",
+		.color = "\033[1;32m",
+		.enabled = 1, .loglevel = LOGL_NOTICE,
+	},
 	[DMM] = {
 		.name = "DMM",
 		.description = "Layer3 Mobility Management (MM)",
@@ -529,6 +551,11 @@ static const struct log_info_cat msc_default_categories[] = {
 		.description = "Supplementary Services",
 		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
+	[DASCI] = {
+		.name = "DASCI",
+		.description = "Advanced Speech Call Items",
+		.enabled = 1, .loglevel = LOGL_NOTICE,
+	},
 };
 
 static int filter_fn(const struct log_context *ctx, struct log_target *tar)
@@ -560,8 +587,12 @@ static int msc_mgw_setup(void)
 	/* Initialize MGW pool. This initalizes and connects all MGCP clients that are currently configured in
 	 * the pool. Adding additional MGCP clients to the pool is possible but the user has to configure and
 	 * (re)connect them manually from the VTY. */
-	pool_members_initalized = mgcp_client_pool_connect(msc_network->mgw.mgw_pool);
-	if (pool_members_initalized) {
+	if (!mgcp_client_pool_empty(msc_network->mgw.mgw_pool)) {
+		pool_members_initalized = mgcp_client_pool_connect(msc_network->mgw.mgw_pool);
+		if (!pool_members_initalized) {
+			LOGP(DMSC, LOGL_ERROR, "MGW pool failed to initialize any pool members\n");
+			return -EINVAL;
+		}
 		LOGP(DMSC, LOGL_NOTICE,
 		     "MGW pool with %u pool members configured, (ignoring MGW configuration in VTY node 'msc').\n",
 		     pool_members_initalized);
@@ -571,15 +602,15 @@ static int msc_mgw_setup(void)
 	/* Initialize and connect a single MGCP client. This MGCP client will appear as the one and only pool
 	 * member if there is no MGW pool configured. */
 	LOGP(DMSC, LOGL_NOTICE, "No MGW pool configured, using MGW configuration in VTY node 'msc'\n");
-	mgcp_client_single = mgcp_client_init(msc_network, &msc_network->mgw.conf);
+	mgcp_client_single = mgcp_client_init(msc_network, msc_network->mgw.conf);
 	if (!mgcp_client_single) {
 		LOGP(DMSC, LOGL_ERROR, "MGW (single) client initalization failed\n");
 		return -EINVAL;
 	}
 	if (mgcp_client_connect(mgcp_client_single)) {
 		LOGP(DMSC, LOGL_ERROR, "MGW (single) connect failed at (%s:%u)\n",
-		     msc_network->mgw.conf.remote_addr,
-		     msc_network->mgw.conf.remote_port);
+		     msc_network->mgw.conf->remote_addr,
+		     msc_network->mgw.conf->remote_port);
 		return -EINVAL;
 	}
 	mgcp_client_pool_register_single(msc_network->mgw.mgw_pool, mgcp_client_single);
@@ -590,6 +621,7 @@ static int msc_mgw_setup(void)
 int main(int argc, char **argv)
 {
 	int rc;
+	int ret = 0;
 
 	struct osmo_sccp_instance *sccp_a;
 	struct osmo_sccp_instance *sccp_iu;
@@ -683,8 +715,7 @@ int main(int argc, char **argv)
 		DEBUGP(DMNCC, "Using internal MNCC handler.\n");
 
 	/* start telnet after reading config for vty_get_bind_addr() */
-	rc = telnet_init_dynif(tall_msc_ctx, &msc_network,
-			       vty_get_bind_addr(), OSMO_VTY_PORT_MSC);
+	rc = telnet_init_default(tall_msc_ctx, &msc_network, OSMO_VTY_PORT_MSC);
 	if (rc < 0)
 		return 2;
 
@@ -742,20 +773,26 @@ TODO: we probably want some of the _net_ ctrl commands from bsc_base_ctrl_cmds_i
 	osmo_init_ignore_signals();
 
 	/* start the SMS queue */
-	if (sms_queue_start(msc_network) != 0)
-		return -1;
+	if (sms_queue_start(msc_network) != 0) {
+		ret = -1;
+		goto error;
+	}
 
-	if (msc_mgw_setup() != 0)
-		return 7;
+	if (msc_mgw_setup() != 0) {
+		ret = 7;
+		goto error;
+	}
 
 	if (ss7_setup(tall_msc_ctx, &sccp_a, &sccp_iu)) {
 		fprintf(stderr, "Setting up SCCP client failed.\n");
-		return 8;
+		ret = 8;
+		goto error;
 	}
 
 	if (sgs_server_open(g_sgs)) {
 		fprintf(stderr, "Starting SGs server failed\n");
-		return 9;
+		ret = 9;
+		goto error;
 	}
 
 	msc_network->a.sri = sccp_ran_init(msc_network, sccp_a, OSMO_SCCP_SSN_BSSAP,
@@ -763,7 +800,8 @@ TODO: we probably want some of the _net_ ctrl commands from bsc_base_ctrl_cmds_i
 					   msc_network);
 	if (!msc_network->a.sri) {
 		fprintf(stderr, "Setting up A receiver failed\n");
-		return 10;
+		ret = 10;
+		goto error;
 	}
 	LOGP(DMSC, LOGL_NOTICE, "A-interface: SCCP user %s, cs7-instance %u (%s)\n",
 	     osmo_sccp_user_name(msc_network->a.sri->scu),
@@ -778,7 +816,8 @@ TODO: we probably want some of the _net_ ctrl commands from bsc_base_ctrl_cmds_i
 					   msc_network);
 	if (!msc_network->iu.sri) {
 		fprintf(stderr, "Setting up IuCS receiver failed\n");
-		return 11;
+		ret = 11;
+		goto error;
 	}
 
 	/* Compatibility with legacy osmo-hnbgw that was unable to properly handle RESET messages. */
@@ -797,7 +836,8 @@ TODO: we probably want some of the _net_ ctrl commands from bsc_base_ctrl_cmds_i
 		rc = osmo_daemonize();
 		if (rc < 0) {
 			perror("Error during daemonize");
-			return 6;
+			ret = 6;
+			goto error;
 		}
 	}
 
@@ -820,6 +860,7 @@ TODO: we probably want some of the _net_ ctrl commands from bsc_base_ctrl_cmds_i
 		}
 	} while (!osmo_select_shutdown_done());
 
+error:
 	db_fini();
 	log_fini();
 
@@ -839,5 +880,5 @@ TODO: we probably want some of the _net_ ctrl commands from bsc_base_ctrl_cmds_i
 	 */
 	talloc_report_full(NULL, stderr);
 	talloc_disable_null_tracking();
-	return 0;
+	return ret;
 }
